


title SWA.ASM  ---  SignWriter Assembly Procedures  ---



subttl   ---  Copyright (c) 1991  Center for Sutton Movement Writing

name SWA

;-------------------------------------------------------------------------------------------------------------------
; File  SWA.ASM
;
;              ---  SignWriter Assembly Procedures  ---
;
;
;    This module provides assembly language routines for the SignWriter
; program.  All external routines/functions are called with a Microsoft
; Pascal interface.
;
;-------------------------------------------------------------------

subttl EQU's, RECORD's, MACRO's, etc.

;-------------------------------------------------------------------
;
;               Equates and Other Definitions
;
;-------------------------------------------------------------------


     ; =======  Miscellaneous Constants  ========================

ChEscape        equ     250     ; Escape code signifying a character
NuSign          equ     253     ; NewSign control code
NuLine          Equ     254     ; NewLine control code
EOF             Equ     255     ; End Of File control code

WorkBytes       equ     256     ; Size of the WorkArea buffer (in bytes)
QSaveBytes      Equ     256     ; Size of QSaveArea buffer
BSaveBytes      Equ     26 * 80 ; Size of BSaveArea buffer (26 lines)

BlockSize       Equ     3       ; Number of pixels in size of zoom box
BlockPad        Equ     1       ; Number of pixels between zoomed boxes

ScanToggle      Equ     200H    ; Bit Xor'd to toggle between odd and even scans
LineSize        Equ     30 * 80 ; # of bytes on even or odd lines in a TL
        ; The following are offsets from B8000 BA000
TL0Start        Equ     0            ; Start offset of even Text Line 0
TL1Start        Equ     31 * 80      ; Start offset of even Text Line 1
TL2Start        Equ     61 * 80      ; Start offset of even Text Line 2
BoxStart        Equ     (100 - 13) * 80 ; Start offset of box area

True            Equ     1
False           Equ     0

VideoInt        Equ     10h     ; ROM BIOS video interrupt
SetGrafMode     Equ     6       ; AH = 0; AL = 6; in call to VideoInt
SetTextMode     Equ     2       ; AH = 0; AL = 2; in call to VideoInt
Bell            Equ     1407h   ; AH = 14; AL = 7; in call to VideoInt
SetCursor       Equ     2       ; AH = 2; DH = row; DL = column

        ; ----- Transformation Constants  ---------

NO_XFORM        equ     0
X_XFORM         equ     1
Y_XFORM         equ     2
XY_XFORM        equ     3

     ; ------ DOS Constants ------

DOSGetKBDVect   Equ     3509h           ; Get the current KBD interrupt vector
DOSSetKBDVect   Equ     2509h           ; Set the current KBD interrupt vector
DOSFlushBuff    Equ     0C00h           ; Flush keyboard buffer
DOSCurDisk      Equ     19h             ; Get current disk
DOSGetDTA       Equ     2Fh             ; DOS func to get the DTA address
DOSGetCurDir    Equ     47h             ; Get current directory
DOSRenameFile   Equ     56h             ; Rename a file
DOSFunction     Equ     21h             ; Interrupt # for DOS functions

     ; ------ Constants used in DrawType -------

DoOr            Equ     0               ; Xor is false
DoXor           Equ     1               ; Xor is true
DoErase         Equ     2               ; Added later for erase

     ; ------ Constants used by AGetChar -------

KLeft	        Equ	08h             ; <Ctrl>-H
KHome	        Equ	88h             ; <Ctrl>-<Alt>-H
KUp	        Equ	0Bh             ; <Ctrl>-K
KPgUp	        Equ	8Bh             ; <Ctrl>-<Alt>-K
KRight	        Equ	0Ch             ; <Ctrl>-L
KEnd	        Equ	8Ch             ; <Ctrl>-<Alt>-L
KDown	        Equ	0Ah             ; <Ctrl>-J
KPgDown	        Equ	8Ah             ; <Ctrl>-<Alt>-J 
KDel            Equ     7Fh             ; Delete

KAltFlag        Equ     10000000b       ; If set, force alt state to true

; Shift state flags
RShiftFlag      Equ     00000001b
LShiftFlag      Equ     00000010b
CtrlFlag        Equ     00000100b
AltFlag         Equ     00001000b

KBDIOInt        Equ     16h             ; ROM BIOS Keyboard interrupt

     ; =======  Bit-Field Records/Masks  ======================

        ;---  Option Record

OptFlds         Record Flop:1, OSize:2, AHead:2, Rotate:3 ;<-- , UNUSED:8


        ;---  Symbol-Attribute Record

SymAttrsEntry   Record SymFlops:1, SymRots:3, SymType:3, SymImgs:3, SymSizes:3


        ;---  Image-File IMG_WID_HD Field

IMG_WIDHD       record  IMG_AHEAD:3, IMG_WIDTH:5

        ;---  Image-File FixPtY and XBias fields

IMG_FixY_XB     record  IMG_XBias:3, IMG_FixPtY:5


     ; =======  Byte-Field Structures  ======================



        ;---  Returned Image Atrs-Record --------------------------
        ;
        ;          ---  this corresponds to the Attrs record defined
        ;               in SW.Pas for return by GETATTRS().
ATRSREC struc

Atrs_SType      DB      ?
Atrs_Sizes      DB      ?
Atrs_Rots       DB      ?
Atrs_Flops      DB      ?
Atrs_Images     DB      ?

ATRSREC ends



        ;---  Returned Image Info-Record --------------------------
        ;
        ;          ---  this corresponds to the Info record defined
        ;               in SW.Pas for return by GETINFO().
INFOREC struc

INFO_XSize      db      ?
INFO_YSize      db      ?
INFO_XFIX       db      ?
INFO_YFIX       db      ?
INFO_XHEAD      db      ?
INFO_YHEAD      db      ?
INFO_ORIENT     db      ?

INFOREC ends


        ;---  Image-File Record --------------------------
        ;
        ;          ---  this corresponds to the Image-File
        ;               record definition (SW-Design 3-87)
IMG_REC struc

IMG_LENGTH      db      ?
IMG_WID_HD      db      ?       ; 2 Bit-Fields!  (Width:5, AHeadOrient:3)
IMG_XHEAD       db      ?
IMG_YHEAD       db      ?
IMG_XFIX        db      ?
ImgYFixXB       db      ?
IMG_IMAGE       db      ?       ; First byte of Image's Bit-Map

IMG_REC ends


FCB_REC         Struc

FCB_Drive       DB      0               ; Current drive
FCB_FileName    DB      '????????'      ; Any file name
FCB_Suffix      DB      'SGN'           ;  - with a .SGN suffix
FCB_Remainder   DB      25 Dup (?)      ; Don't care what's in the rest of it

FCB_REC         Ends


;-------------------------------------------------------------------
;
;               MACRO Definitions
;
;-------------------------------------------------------------------

.sall   ; suppress Macro Expansion Listing


STARTPROC       macro   proc_name

                public  proc_name
proc_name       proc    far
                push    bp
                mov     bp,sp
                endm
; -------------------------------------

ENDPROC         macro   proc_name,param_bytes

                pop     bp
                ret     param_bytes
proc_name       endp
                endm
; -------------------------------------


subttl      ---  Global Data  ---
page

;-------------------------------------------------------------------
;
;                       Global Data
;
;-------------------------------------------------------------------

DGROUP  group _DATA

        assume  CS:SW_Supp, DS:DGROUP, ES:DGROUP, SS:DGROUP


_DATA   segment word public     'DATA'

                ; ----- External Tables -----

Extrn           TSymIndex:DWord
Extrn           TChrIndex:DWord
Extrn           TSymAttr: DWord

                ; -----  Screen Buffer Constants  ---------

CGAEvenScans    DW      0B800h  ; Value for SegReg of buffer for even scan lines
CGAOddScans     DW      0BA00h  ; Value for SegReg of buffer for odd scan lines
BytesPerRow     DW      80      ; # of bytes per row in screen buffer

        ; Address of 1st even scan in each text line
LineEAddrs      DW      TL0Start        ; Seg:Off address of TextLine0
                DW      TL1Start        ; Seg:Off address of TextLine1
                DW      TL2Start        ; Seg:Off address of TextLine2

        ; Address of 1st odd scan in each text line
LineOAddrs      DW      TL0Start        ; Seg:Off address of TextLine0
                DW      TL1Start - 80   ; Seg:Off address of TextLine1
                DW      TL2Start        ; Seg:Off address of TextLine2

        ; < Public and Extern data declarations go here >

CHAR            DW      0               ; Current Symbol/Character Code
OPTION          DW      0               ; Currrent Option Code
DrawType        DB      0               ; Determines if SYMBOL_BLT XOR's

public CHAR, OPTION             ; --- For DEBUG Only ----


XPOS            dW      ?               ; this is a fullscale X,Y coordinate
XPOS2           dW      ?
YPOS            dW      ?               ; this is a fullscale X,Y coordinate
XSize           db      ?
YSize           db      ?
XBias           DW      ?               ; Unused bits in rightmost byte of image.

QD_XPOS        dW      ?               ; XPos for QDraw
QD_YPOS        dW      ?               ; YPos for QDraw
QD_XSize       db      ?               ; XSize for QDraw
QD_YSize       db      ?               ; YSize for QDraw

public  YSize, XSize, YPOS, XPOS        ; --- For DEBUG Only ----

        ;------ Temps used by GETINFO() ----------

TRANS           db      0               ;set by LGetImg()

ImageLen        db      ?
XFIX            db      ?
YFIX            db      ?
XHEAD           db      ?
YHEAD           db      ?
AHOrient        db      ?

public TRANS, ImageLen, XFIX              ; --- For DEBUG Only ----
public YFIX,XHEAD,YHEAD,AHOrient       ; --- For DEBUG Only ----



        ;------ Variables used by LookupAttrs() ----------

public  ATS  ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
ATS                     equ     this byte

AttrSizes               db      ?
AttrImages              db      ?
AttrSType               db      ?
AttrRots                db      ?
AttrFlops               DB      ?

public  AttrSizes, AttrImages ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  AttrSType, AttrRots  ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

        ;------ Variables used by TRANSLATE_OPTION() ----------

public  OPTS  ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
OPTS                    equ     this byte

OPT_ROTATION            db      ?
OPT_AHEAD               db      ?
OPT_SIZE                db      ?
OPT_FLOP                db      ?

public  OPT_ROTATION, OPT_AHEAD ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  OPT_SIZE, OPT_FLOP      ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

        ; ---  Other General Pointers  -------------

ImgPtr          DD      ?               ;^to base of Image Record

public ImgPtr           ; --- For DEBUG Only ----


; ------  Temporary Buffers  ---------

Public WorkArea                         ; --- For DEBUG Only ----
WORKAREA        db      WorkBytes dup (?)

Public QSaveArea                        ; --- For DEBUG Only ----
QSaveArea       DB      QSaveBytes Dup (?) ; Save area for QDraw

Public BSaveArea                        ; --- For DEBUG Only ----
BSaveArea       DB      BSaveBytes Dup (?) ; Save area for SaveBoxArea

; public FLIP                           ; --- For DEBUG Only ----
; FLIP          db      256 dup (55h)   ; 55 is "uninitialized" value


; Variables used by KBD_Int and AGetChar

Old_KBD_Int     DD      ?               ; Address of the old interrupt routine
Public  Old_KBD_Int
CBuffSize       Equ     16              ; Size of character buffer
; Circular buffers of incoming chars
CharBuff        DB      CBuffSize Dup (?) ; AL from KBD BIOS, AH = 0
SpecBuff        DB      CBuffSize Dup (?) ; AH from KBD BIOS, AH = 0
ShiftBuff       DB      CBuffSize Dup (?) ; AL from KBD BIOS, AH = 2
; Mapping table from special codes to characters
CharMap         DB      16 Dup (0)      ; 16 unused nulls
                DB      "QWERTYUIOP"    ; Alt-<letter> -> <letter>
                DB      4 Dup (0)       ; 4 unused nulls
                DB      "ASDFGHJKL"     ; Alt-<letter> -> <letter>
                DB      5 Dup (0)       ; 5 unused nulls
                DB      "ZXCVBNM"       ; Alt-<letter> -> <letter>
                DB      20 Dup (0)      ; 20 unused nulls
                DB      KHome,KUp,KPgUp,0,KLeft,0
                DB      KRight,0,KEnd,KDown,KPgDown,0,KDel
                DB      47 Dup (0)      ; 20 unused nulls
                DB      '+',0           ; <Alt>+, unused null
; Indexes into the buffers
NextCh          DW      CBuffsize - 1   ; Index into ChBuffer of next char
LastCh          DW      0               ; Index into ChBuffer of last char

page


; AHMAP translates arrowhead orientation values (0..7) according
; to the image transform value (0..3).

AHMAP           db      0,0,4,4
                db      1,7,3,5
                db      2,6,2,6
                db      3,5,1,7
                db      4,4,0,0
                db      5,3,7,1
                db      6,2,6,2
                db      7,1,5,3


;-------------------------------------------------------------------------
;
; The following tables map rotation values (in the range 0..MaxRot) into
; the corresponding image number (0..MaxImg) and transform value for each
; of the five symbol classes.  Digits in the table names indicate the
; number of images in the symbol class.  Tables named --Fx yield
; analogous values for flopped images.
;
; Unused table entries are defined with 0 values to provide a degree of
; robustness in case of invalid rotation values.  (NOTE - Do not change
; or reorder these declarations without studying the table lookup code
; above.)
;
; Transform values are defined as follows:
;
;       0    no transform
;       1    X  transform
;       2    Y  transform
;       3    XY transform
;


IM              db      0, 0, 0, 0, 0, 0, 0, 0
IM2             db      0, 1, 1, 0, 0, 0, 0, 0
IM3             db      0, 1, 2, 1, 0, 1, 2, 1
IM4             db      0, 1, 2, 3, 0, 1, 2, 3
IM5             db      0, 1, 2, 3, 4, 3, 2, 1

TR              db      0, 2, 0, 0, 0, 0, 0, 0
TR2             db      0, 0, 2, 0, 0, 0, 0, 0
TR3             db      0, 0, 0, 1, 1, 3, 2, 2
TR4             db      0, 0, 0, 0, 3, 3, 3, 3
TR5             db      0, 0, 0, 0, 0, 1, 1, 1


IMF             db      0, 0, 0, 0, 0, 0, 0, 0
IMF2            db      0, 1, 1, 0, 0, 0, 0, 0
IMF3            db      0, 1, 2, 0, 0, 0, 0, 0
IMF4            db      0, 3, 2, 1, 0, 3, 2, 1
IMF5            db      0, 1, 2, 3, 4, 3, 2, 1

TRF             db      1, 3, 0, 0, 0, 0, 0, 0
TRF2            db      1, 1, 3, 0, 0, 0, 0, 0
TRF3            db      1, 1, 1, 0, 0, 0, 0, 0
TRF4            db      1, 2, 2, 2, 2, 1, 1, 1
TRF5            db      0, 0, 0, 0, 0, 1, 1, 1


public  IM,TR, IMF,TRF  ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

page

;-------------------------------------------------------------------------
;
; The following table maps byte values to their X-wise mirror image

public  FLIP ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

FLIP            db      00000000B, 10000000B, 01000000B, 11000000B
                db      00100000B, 10100000B, 01100000B, 11100000B

                db      00010000B, 10010000B, 01010000B, 11010000B
                db      00110000B, 10110000B, 01110000B, 11110000B

                db      00001000B, 10001000B, 01001000B, 11001000B
                db      00101000B, 10101000B, 01101000B, 11101000B

                db      00011000B, 10011000B, 01011000B, 11011000B
                db      00111000B, 10111000B, 01111000B, 11111000B

                db      00000100B, 10000100B, 01000100B, 11000100B
                db      00100100B, 10100100B, 01100100B, 11100100B

                db      00010100B, 10010100B, 01010100B, 11010100B
                db      00110100B, 10110100B, 01110100B, 11110100B

                db      00001100B, 10001100B, 01001100B, 11001100B
                db      00101100B, 10101100B, 01101100B, 11101100B

                db      00011100B, 10011100B, 01011100B, 11011100B
                db      00111100B, 10111100B, 01111100B, 11111100B

                db      00000010B, 10000010B, 01000010B, 11000010B
                db      00100010B, 10100010B, 01100010B, 11100010B

                db      00010010B, 10010010B, 01010010B, 11010010B
                db      00110010B, 10110010B, 01110010B, 11110010B

                db      00001010B, 10001010B, 01001010B, 11001010B
                db      00101010B, 10101010B, 01101010B, 11101010B

                db      00011010B, 10011010B, 01011010B, 11011010B
                db      00111010B, 10111010B, 01111010B, 11111010B

                db      00000110B, 10000110B, 01000110B, 11000110B
                db      00100110B, 10100110B, 01100110B, 11100110B

                db      00010110B, 10010110B, 01010110B, 11010110B
                db      00110110B, 10110110B, 01110110B, 11110110B

                db      00001110B, 10001110B, 01001110B, 11001110B
                db      00101110B, 10101110B, 01101110B, 11101110B

                db      00011110B, 10011110B, 01011110B, 11011110B
                db      00111110B, 10111110B, 01111110B, 11111110B

                db      00000001B, 10000001B, 01000001B, 11000001B
                db      00100001B, 10100001B, 01100001B, 11100001B

                db      00010001B, 10010001B, 01010001B, 11010001B
                db      00110001B, 10110001B, 01110001B, 11110001B

                db      00001001B, 10001001B, 01001001B, 11001001B
                db      00101001B, 10101001B, 01101001B, 11101001B

                db      00011001B, 10011001B, 01011001B, 11011001B
                db      00111001B, 10111001B, 01111001B, 11111001B

                db      00000101B, 10000101B, 01000101B, 11000101B
                db      00100101B, 10100101B, 01100101B, 11100101B

                db      00010101B, 10010101B, 01010101B, 11010101B
                db      00110101B, 10110101B, 01110101B, 11110101B

                db      00001101B, 10001101B, 01001101B, 11001101B
                db      00101101B, 10101101B, 01101101B, 11101101B

                db      00011101B, 10011101B, 01011101B, 11011101B
                db      00111101B, 10111101B, 01111101B, 11111101B

                db      00000011B, 10000011B, 01000011B, 11000011B
                db      00100011B, 10100011B, 01100011B, 11100011B

                db      00010011B, 10010011B, 01010011B, 11010011B
                db      00110011B, 10110011B, 01110011B, 11110011B

                db      00001011B, 10001011B, 01001011B, 11001011B
                db      00101011B, 10101011B, 01101011B, 11101011B

                db      00011011B, 10011011B, 01011011B, 11011011B
                db      00111011B, 10111011B, 01111011B, 11111011B

                db      00000111B, 10000111B, 01000111B, 11000111B
                db      00100111B, 10100111B, 01100111B, 11100111B

                db      00010111B, 10010111B, 01010111B, 11010111B
                db      00110111B, 10110111B, 01110111B, 11110111B

                db      00001111B, 10001111B, 01001111B, 11001111B
                db      00101111B, 10101111B, 01101111B, 11101111B

                db      00011111B, 10011111B, 01011111B, 11011111B
                db      00111111B, 10111111B, 01111111B, 11111111B

DL_Masks        DB      8 Dup (0)
DL_SolidMasks   DB      11111111B, 01111111B, 00111111B, 00011111B
                DB      00001111B, 00000111B, 00000011B, 00000001B
DL_DotMasks     DB      10101010B, 00101010B, 00101010B, 00001010B
                DB      00001010B, 00000010B, 00000010B, 00000000B
DL_NotMasks     DB      00000000B, 10000000B, 11000000B, 11100000B
                DB      11110000B, 11111000B, 11111100B, 11111110B

; Table used by AGetChar to convert from extended function keys to
; an ASCII equivalent.

EXFCodeToAscii  DB      15 Dup (0)
                DB      "QWERTYUIOP"
                DB      4 Dup (0)
                DB      "ASDFGHJKL"
                DB      5 Dup (0)
                DB      "ZXCVBNM"
                DB      64 Dup (0)


_DATA   ends            ; ======== End of the Data Segment  =============

SW_Supp segment 'CODE'    ; ======= Start of the Code Segment  ==========

subttl ---                                              Interrupt: KBD_Int
page
;-------------------------------------------------------------------
;
;
; procedure KBD_Int
;
;       KBD_Int intercepts the keyboard interrupt, calls the ROM BIOS keyboard
;       routine, and then stores the resulting key in a buffer.
;
Public KBD_Int
KBD_Int         Proc    Far
        Push    DS
        Push    AX
        Push    BX
        Push    CX        

        Mov     AX,DGroup
        Mov     DS,AX
        Pushf                           ; Push flags to simulate an interrupt
        Call    Old_KBD_Int             ; Let the old interrupt process the char
        Sti

        Mov     AH,1
        Int     KBDIOInt                ; Is a character ready for us?
        Jz      KBDDone

        Xor     AH,AH                   ; Yes,
        Int     KBDIOInt                ; Get next char in buffer
        Mov     CX,AX                   ; Save the code in CX
        Mov     AH,2
        Int     KBDIOInt                ; Quick, get the shift state before
                                        ; -it changes.

        Mov     BX,LastCh               ; Get index and update it.
        Cmp     BX,NextCh               ; Is there room left in the buffer?
        Je      KBDBeep

        Mov     CharBuff[BX],CL         ; Yes, store the char
        Mov     SpecBuff[BX],CH         ; Store the special code
        Mov     ShiftBuff[BX],AL        ; Store the shift state

        Inc     BX                      ; Update the LastCh pointer
        And     BX,CBuffSize - 1
        Mov     LastCh,BX

        Jmp     Short KBDDone

KBDBeep:
        Mov     AX,Bell                 ; 14:07 service:char
        Int     VideoInt                ; Ri-i-ing ring my bell.

KBDDone:
        Pop     CX
        Pop     BX
        Pop     AX
        Pop     DS
        Iret                            ; Return from the interrupt

KBD_Int         EndP


subttl ---                                              Procedure:  AsmInit
page
;-------------------------------------------------------------------
;
;
; Procedure AsmInit;
;
;       Does initialization for the assembly module.  Currently it just
;       sets up the keyboard interrupt.
;
StartProc       AsmInit

AI_ClrKBD:                              ; Clear the current keyboard buffer
        Mov     AX,DOSFlushBuff
        Int     DOSFunction

AI_SwapInt:                             ; Replace the keyboard interrupt
        Mov     AX,DOSGetKBDVect        ; Get the old vector
        Int     DOSFunction
        Mov     Word Ptr Old_KBD_Int,BX ; Store it in Old_KBD_Int
        Mov     Word Ptr Old_KBD_Int + 2,ES

        Mov     AX,DOSSetKBDVect        ; Set the new vector
        Mov     DX,Offset KBD_Int
        Push    DS
        Mov     BX,CS
        Mov     DS,BX
        Int     DOSFunction
        Pop     DS

EndProc         AsmInit
        
subttl ---                                              Procedure:  AsmTerm
page
;-------------------------------------------------------------------
;
;
; Procedure AsmTerm;
;
;       Does termination for the assembly module.  Currently it just
;       resets the keyboard interrupt.
;
StartProc       AsmTerm

        Push    DS
        Lds     DX,Old_KBD_Int
        Mov     AX,DOSSetKBDVect
        Int     DOSFunction
        Pop     DS

EndProc         AsmTerm
        
subttl ---                                              Procedure:  GrafMode
page
;-------------------------------------------------------------------
;
;
; procedure GrafMode;
;
;       GrafMode sets up the screen for high density, monochrome graphics.
;
StartProc       GrafMode

        Mov     AX,SetGrafMode
        Int     VideoInt

EndProc         GrafMode

subttl ---                                              Procedure:  TextMode
page
;-------------------------------------------------------------------
;
;
; procedure TextMode;
;
;       TextMode sets up the screen for text mode
;
;-------------------------------------------------------------------
StartProc       TextMode

        Mov     AX,SetTextMode
        Int     VideoInt

EndProc         TextMode

subttl ---                                              Procedure:  ClearScreen
page
;-------------------------------------------------------------------
;
;
; procedure ClearScreen;
;
;       ClearScreen clears the screen after GrafMode has been set up.
;
StartProc       ClearScreen

        Mov     ES,CGAEvenScans
        Xor     DI,DI
        Cld
        Mov     CX,4000         ; # of words per half screen
        Xor     AX,AX
        Rep     Stosw

        Mov     ES,CGAOddScans
        Xor     DI,DI
        Mov     CX,4000         ; # of words per half screen
        Rep     Stosw

EndProc         ClearScreen


subttl ---                                              Sub-Routine:  TRANSLATE_OPTION
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine TRANSLATE_OPTION
;
;
;
;    The TRANSLATE_OPTION unpacks the global OPTION variable's Bit-
; Fields into the global variables OPT_ROTATION, OPT_AHEAD, OPT_SIZE
; and OPT_FLOP, using the OPT_FLDS      record definition.
;
;
; WARNING:  No registers are preserved.
;
;-------------------------------------------------------------------

public   TRANSLATE_OPTION ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

TRANSLATE_OPTION        proc    near


        mov     bx,OPTION               ;load the Option, set previously by caller


        ; --- Isolate the ROTATE Bit-Field ----------

        mov     ax,bx
        and     ax,mask ROTATE
        mov     cl,ROTATE
        shr     ax,cl
        mov     OPT_ROTATION,al


        ; --- Isolate the AHEAD Bit-Field ----------

        mov     ax,bx
        and     ax,mask AHEAD
        mov     cl,AHEAD
        shr     ax,cl
        mov     OPT_AHEAD,al


        ; --- Isolate the OSize Bit-Field ----------

        mov     ax,bx
        and     ax,mask OSIZE
        mov     cl,OSIZE
        shr     ax,cl
        mov     OPT_SIZE,al


        ; --- Isolate the FLOP Bit-Field ----------

        mov     ax,bx
        and     ax,mask FLOP
        mov     cl,FLOP
        shr     ax,cl
        mov     OPT_FLOP,al


        ; --- Adjust Size value by Ahead bias ----------

	mov	al,OPT_AHEAD
	cmp	al,0
	jne	TRANS_END

	mov	al,OPT_SIZE
	add	al,4
	mov	OPT_SIZE,al


TRANS_END:

        ret

TRANSLATE_OPTION        endp

subttl ---                                              Sub-Routine:  LookupAttrs
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine LookupAttrs
;
;
;
;    The LookupAttrs routine looks up the Attribute Byte correspond-
; ing to the global CHAR code.  The Attribute's Bit-Fields are
; isolated and stuck in the AttrSizes, AttrImages, and AttrSType
; global variables.  The AttrImages result is then used to index
; the global Rotations Table (NROTS), the result of which is placed
; in the AttrRots global variable.
;
;
; WARNING:  No registers are preserved.
;
;-------------------------------------------------------------------

public   LookupAttrs ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

LookupAttrs    proc    near


        ; index into the symbol attribute record

        Mov     AX,Char                 ;load CHAR for use as index
        Shl     AX,1                    ; to make it a word-index
        Les     BX,TSymAttr             ; ES:BX -> TSymAttr[0]
        Add     BX,AX                   ; ES:BX -> TSymAttr[Char]
        mov     DX,ES:[BX]              ;lookup the CHAR's Attribute Byte

        Mov     AX,DX                   ; Get SymAttr
        and     AX,mask SymSizes        ; Isolate the OSIZES Bit-Field
        mov     AttrSizes,al            ;   and save it

        mov     ax,DX                   ; Get SymAttr
        and     ax,mask SymImgs         ; Isolate the IMAGES Bit-Field
        mov     cl,SymImgs
        shr     ax,cl
        mov     AttrImages,al           ;   and save it

        mov     ax,DX                   ; Get SymAttr
        and     ax,mask SymType         ; Isolate the SYMBOL-TYPE Bit-Field
        mov     cx,SymType
        shr     ax,cl
        mov     AttrSType,al            ;   and save it

        mov     ax,DX                   ; Get SymAttr
        and     ax,mask SymRots         ; Isolate the SYMBOL-ROTS Bit-Field
        mov     cx,SymRots
        shr     ax,cl
        mov     AttrRots,al             ;   and save it

        mov     ax,DX                   ; Get SymAttr
        and     ax,mask SymFlops        ; Isolate the SYMBOL-FLOPS Bit-Field
        mov     cx,SymFlops
        shr     ax,cl
        mov     AttrFlops,al            ;   and save it

        ret

LookupAttrs    endp

subttl ---                                              Routine:  GETATTRS
page
;-------------------------------------------------------------------
;
; procedure GetAttrs( CHAR_VAL: byte; var ATTR_RECORD: Attrs );
;
;
;    The GETATTRS routine returns a structure of type ATTR_REC via
; the ATTR_RECORD parameter, with the Attribute Record corresponding
; to the CHAR_VAL symbol passed.
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------

_CHAR_VAL       =  8
_ATTR_PTR       =  6


STARTPROC       GETATTRS


        ;--- Load Character argument into global CHAR  ------


        mov     ax,_CHAR_VAL[bp]
        xor     ah,ah                   ;Limit CHAR to <= 255.
        mov     CHAR,ax


        ;--- Look up CHAR_VAL's Attribute byte

        call    LookupAttrs



        ;--- Return the values requested

        mov     bx,_ATTR_PTR[bp]        ;get address of record to set

        mov     al,AttrSizes
        mov     [bx].Atrs_Sizes,AL

        mov     al,AttrSType
        mov     [bx].Atrs_SType,AL

        mov     al,AttrRots
        mov     [bx].Atrs_Rots,AL

        mov     al,AttrFlops
        mov     [bx].Atrs_Flops,AL

        mov     al,AttrImages
        mov     [bx].Atrs_Images,AL

ENDPROC         GETATTRS, 4             ;return and pop 4 bytes


subttl ---                                              Sub-Routine:  TRAVERSE
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine TRAVERSE
;
;
;   Traverses the currently addressed images to get the correct one,
; based on the Size requested in the DX-reg.
;                                   ------
;
; WARNING:  No registers are preserved.
;
;-------------------------------------------------------------------

public  TRAVERSE ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

TRAVERSE        proc    near


;TRAVERSE                                ;moves imgptr past symbol images


        xor     dh,dh
        cmp     dl,0
        je      TRAV_BAIL


        xor     ax,ax                   ;0 the uppper byte for addition
        Les     bx,ImgPtr               ;Load pointer to the Image


TRAV_AGAIN:     ;---- Skip over successive records, using each Img-Length
                ;       byte to tell how far to jump.  Keep doing this
                ;       until image-count passed (DL) is decremented to 0.

        mov     al,ES:[bx].IMG_LENGTH      ;Get current record's image-length

        add     bx,ax                   ;Add length to the current pointer
        add     bx,6                    ;Skip the record-header too

        dec     dl                      ;Traversed enough images yet?
        jnz     TRAV_AGAIN              ;   If not, do it again

        mov     Word Ptr ImgPtr,bx       ;save final result as new Image^

TRAV_BAIL:

        ret

TRAVERSE        endp


subttl ---                                              Routine:  LGetImg
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine LGetImg
;
;
;   "code+option -> imgptr, transval"   (ala Rich)
;
;        In other words:
;
;
;    The LGetImg routine establishes the global ImgPtr to point at
; the image data based on the global CHAR and OPTION values.
;
;-------------------------------------------------------------------

public  LGetImg         ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=


LGetImg proc    near


        mov     DI,CHAR                 ;character symbol code?
        cmp     DI,ChEscape
        jne     NOT_ChEscape

        mov     DI,OPTION               ; use Option as symbol code
        Les     BX,TChrIndex            ; Get address of TCharIndex table
        mov     OPTION, 00001000B       ; stock option byte
        jmp     CALC_IMGADR

NOT_ChEscape:           ; ---- Not an Escape-Code  -----------

        Les     BX,TSymIndex            ; Get address of TSymIndex table


CALC_IMGADR:     ; --- Get base address for the specified image size

        shl     DI,1                    ; Convert Code to segptr index
        shl     DI,1
        Les     BX,ES:[BX+DI]           ; get image offset
        mov     Word Ptr ImgPtr,BX
        mov     Word Ptr ImgPtr + 2,ES


   ; --- Set up global Attribute and Option Information  -----

        call    LookupAttrs            ;Attributes from Attribute-Table --> ATTR_*

        call    TRANSLATE_OPTION       ;Translate OPTION's Bit-Fields --> OPT_*

        xor     dh,dh
        mov     dl,OPT_SIZE     ; move req'd size into DX

  ;---- Error Check the Size parameter -------
        cmp     dl,AttrSizes
        jle     SMALL_ENOUGH

        mov     dl,AttrSizes
        mov     OPT_SIZE,DL     ;fix the parameter value

SMALL_ENOUGH:   ;--- Passed the MAXimum Check

        mov     al,AttrImages   ; dl := #images * size
        mul     dl
        mov     dx,ax 

        call    TRAVERSE        ;skip to proper image



   ; ----- Get transform value, Traverse to proper transformed image ------

        ;--- Calculate the Transformation Table offset by multiplying the
        ;       (number of images - 1) * the width of the table(8), then
        ;       adding in the desired rotation.

public  CALC_XFORM      ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public   IMG_IS_FLOPPED ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  DO_XVERSAL      ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
CALC_XFORM:             ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

        xor     bx,bx
        mov     bl,AttrImages
        dec     bl
        shl     bl,1
        shl     bl,1
        shl     bl,1            ;BL = (AttrImages - 1) * 8

        add     bl,OPT_ROTATION ;BL = Rotation-Adjusted Offset into Table Row


        mov     dx,0            ;0-out this counter used later

        mov     al,OPT_FLOP
        cmp     al,1            ;Is the Image FLOPPED?
        je      IMG_IS_FLOPPED


                                ;  NO, it's NOT flopped ...
                                ;
        mov     al,TR[bx]       ;    TRANS =  TR[bx]
        mov     TRANS,al        ;
        mov     dl,IM[bx]       ;    DL = image #  (passed to TRAVERSE below)

        jmp     DO_XVERSAL



IMG_IS_FLOPPED:                 ;  YES it's flopped ...
                                ;
        mov     al,TRF[bx]      ;    TRANS =  TRF[bx]
        mov     TRANS,al        ;
        mov     dl,IMF[bx]      ;    DL = image #  (passed to TRAVERSE below)



DO_XVERSAL:

        call    TRAVERSE                ;skip to proper image


        ;get image X- & Y-size

        Les     BX,ImgPtr

                                        ;get X-size (in bytes)
        mov     al,ES:[BX].IMG_WID_HD
        and     al,mask IMG_WIDTH               ;old: "AND     #00011111T"
        mov     XSize,al

        mov     al,ES:[BX].IMG_LENGTH              ;get # image bytes
        mov     ImageLen,al

        Xor     AH,AH                           ; Calc. YSize
        Div     XSize
        Mov     YSize,AL

        mov     al,ES:[BX].ImgYFixXB               ; Get X-Bias
        Mov     CL,IMG_XBias
        Shr     AL,CL
        Xor     AH,AH                   ; convert to a word.
        Mov     XBias,AX
         
       ret

LGetImg endp


subttl ---                                              Procedure:  GETINFO
page
;-------------------------------------------------------------------
;
; procedure GETINFO( CHAR_VAL, OPTION: byte; var INFO_REC );
;
;
;    The GETINFO routine returns a structure of type INFOREC via
; the INFO_REC parameter, with the Image information of the CHAR_VAL
; symbol and OPTION value.
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------

_CHAR_VAL       = 10
_OPTION         =  8
_INFO_REC       =  6


public  GETINFO         ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=


STARTPROC       GETINFO

;       .PROC   GETINFO,3               ; returns image info
;       .REF    LGetImg


        ;--- Get arguments into globals OPTION and CHAR  ------

        mov     ax,_OPTION[bp]
        xor     ah,ah                   ;Limit the OPTION to 255
        mov     OPTION,ax

        mov     ax,_CHAR_VAL[bp]
        xor     ah,ah                   ;Limit the CHAR to 255
        mov     CHAR,ax


        call    LGetImg         ;get image info rec

        ;---- Load BX to point at the base of the Image-File Data for CHAR

        Mov     AL,XSize
        Shl     AL,1
        Shl     AL,1
        Shl     AL,1
        Sub     AL,Byte Ptr XBias
        Mov     XSize,AL

        Les     bx,ImgPtr       ;Load ^to the Image-Record (set by LGetImg)

        xor     ax,ax
        mov     al,ES:[BX].IMG_WID_HD      ;get (AHeadOrient * 4)
        and     al,mask IMG_AHEAD
        mov     cl,(IMG_AHEAD - 2)      ; (-2) as we are also mult'ing by 4!
        shr     al,cl
        add     al,TRANS                ;add image transform val
        mov     SI,ax
        mov     al,AHMAP[SI]            ;get new AH orient value
        mov     AHOrient,AL             ;save in AHOrient

        ;--- Get fix and arrowhead point values  ----

        mov     al,ES:[BX].IMG_XHEAD
        mov     XHEAD,al

        mov     al,ES:[BX].IMG_YHEAD
        mov     YHEAD,al

        mov     al,ES:[BX].IMG_XFIX
        mov     XFIX,al

        mov     al,ES:[BX].ImgYFixXB
        And     AL,Mask IMG_FixPtY
        mov     YFIX,al


        ;map fix and arrowhead point values

        Mov     AL,Trans                ; Get transform value
        Test    AL,1                    ; Is it even or odd?
        Jz      TestYForm               ; If even then no X transformation

        Mov     AL,XSize                ; XHead := XSize - XHead - 1
        Sub     AL,XHead
        Dec     AL
        Mov     XHead,AL

        Mov     AL,XSize                ; XFix := XSize - XFix - 1
        Sub     AL,XFix
        Dec     AL
        Mov     XFix,AL

TestYForm:
        Mov     AL,Trans                ; Get transform value
        Cmp     AL,2                    ; >= 2 means do Y transformation
        Jl      GIReturnVals            ; If not, just store return values

        Mov     AL,YSize                ; YHead := YSize - YHead - 1
        Sub     AL,YHead
        Dec     AL
        Mov     YHead,AL

        Mov     AL,YSize                ; YFix := YSize - YFix - 1
        Sub     AL,YFix
        Dec     AL
        Mov     YFix,AL

        ;---- Store info into return record -------

GIReturnVals:
        mov     bx,_INFO_REC[bp]        ;get address of return record

        Mov     AL,XSize
        mov     [bx].INFO_XSize,AL

        mov     al,YSize
        mov     [bx].INFO_YSize,AL

        mov     al,XFIX
        And     AL,11111110B            ; Clear low bit
        mov     [bx].INFO_XFIX,AL

        mov     al,YFIX
        mov     [bx].INFO_YFIX,AL

        mov     al,XHEAD
        mov     [bx].INFO_XHEAD,AL

        mov     al,YHEAD
        mov     [bx].INFO_YHEAD,AL

        mov     AL,AHOrient         ; ahead orient   (set above!)
        mov     [bx].INFO_ORIENT,AL

ENDPROC         GETINFO, 6              ;return and pop 6 bytes



subttl ---                                              Sub-Routine:  Y_XFORM_ROW
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine Y_XFORM_ROW
;
;
;
;    Y_XFORM_ROW performs a Y-plane transformation of the row currently
; being transferred, as indicated by the SI-register, which is assumed
; to be pointing to the begin of some symbol image-row.  The row is
; copied into WORKAREA row indicated by the DI-register, and flipped
; simultaneously.
;
;
; INPUT
;       DS:SI-reg   - ^to Image Row (source of data to be flipped)
;       ES:DI-reg   - ^to start of Destination (WORKAREA) row
;       XSize    - The width of each row to be drawn (in bytes)
;
; OUTPUT
;       DS:SI-reg   - ^to byte past the end of the Source-Row
;       ES:DI-reg   -  "   "    "    "   "  "   "  Destination-Row
;
;
; WARNING:  No registers are preserved.
;           DS does not point to DGroup
;
;-------------------------------------------------------------------

public  Y_XFORM_ROW     ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

Y_XFORM_ROW     proc    near
        push    bx                              ; Save reg's
        push    cx

        ;--- Start at end of the Source-Row

        xor     cx,cx
        mov     cl,ES:XSize
        add     si,cx
        dec     si

ROW_YF_LOOP:
        xor     bh,bh
        mov     bl,[si]                         ; Get old image
        mov     al,ES:FLIP[bx]                  ; Flip it
        Stosb                                   ; Store at start of Dest-Row
        dec     si                              ; Decrement source pointer
        loop    ROW_YF_LOOP                     ; Continue til end of row

        ;--- Move SI past the end of the source row for return

        xor     cx,cx
        mov     cl,ES:XSize
        add     si,cx
        inc     si

        pop     cx                              ; Restore saved reg's
        pop     bx
        ret

Y_XFORM_ROW     endp

subttl ---                                              Routine:  LSetup
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine LSetup
;
;
;    The LSetup routine sets up a symbol for display to the screen.
; The CHAR symbol with the OPTION display option is prepared for
; display, with its upper-left corner at the screen location (XPOS,
; YPOS).  The arguments are pulled write off of the callers argument
; stack.  The stack should look like this:
;
;       Code
;       Option
;       X
;       Y
;       CS
;       IP   (of return address in Pascal module)
;   BP->OldBP
;   SP->IP   (of return address in Draw, Erase or QDraw)
;
;  ------ Some other stuff is done and returned.  Film at 11. ------
;
;
; WARNING:  No registers are preserved.
;
; ---  Parameter Offsets -----------

_CHAR_VAL       = 12
_OPTION         = 10
_XPOS           =  8
_YPOS           =  6

;-------------------------------------------------------------------

public  LSetup  ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  NOFORM  ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  XFORM   ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  YFORM   ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  XYFORM  ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  ROW_MLOOP ;=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
public  YFORM      ;=-=-=-=-=-=  DEBUG Only   -=-=-=-=-=-=-=
public  CONTINUE   ;=-=-=-=-=-=  DEBUG Only   -=-=-=-=-=-=-=


LSetup  proc    near

        Push    DS                      ; DS gets lost later

        ;--- Get the arguments into corresponding globals  ------

        mov     ax,_OPTION[bp]
        xor     ah,ah                   ;Limit the OPTION to 255
        mov     Option,ax

        mov     ax,_CHAR_VAL[bp]
        xor     ah,ah                   ;Limit the CHAR to 255
        mov     Char,ax

        mov     ax,_XPOS[bp]
        mov     XPos,ax

        mov     ax,_YPOS[bp]
        mov     YPos,ax

        ;--- Get image's X- & Y-size from image info ---

        call    LGetImg         ;get image info rec address

        ;---- Move image pointer past image info rec

        Add     Word Ptr ImgPtr,6

        ;---- Invoke appropriate transformation algorithm

public  XFORM_JUMP      ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=
XFORM_JUMP:             ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

        ; Set up pointers for transformation

        Mov     AX,DS
        Mov     ES,AX           ; ES := DS
        Mov     DI,Offset DGroup:WorkArea       ;  ES:DI -> WorkArea
        Lds     SI,ImgPtr       ; DS:SI -> bit map of image in image table
        Cld                     ; Clear for all the following code.

        mov     ah,ES:Trans     ;Use Transformation value set by LGetImg

        cmp     ah,X_XFORM
        je      XFORM

        cmp     ah,Y_XFORM
        je      YFORM

        cmp     ah,XY_XFORM
        je      XYFORM


        jmp     far ptr NOFORM

XFORM:  ; ---- Perform X-plane inversion  ----------------

        mov     dl,ES:YSize             ;DL-reg :=  YSize

ROW_XLOOP:
        call    Y_XFORM_ROW             ;Xfer and FLIP the row
        mov     al,0                    ;slap a zero on end of the row
        Stosb                           ;[di] :=  0;  (di++)
        dec     dl
        jg      ROW_XLOOP               ;done all the rows?

        jmp     far ptr CONTINUE        ; Jump to the end of the Transformations

YFORM:  ;---- Perform Y-plane inversion  -----------------

        Mov     ES:XBias,0              ; XBias only used on XForm

        ;--- Start with the LAST ROW of the Image ---------

        xor     ax,ax
        mov     al,ES:ImageLen
        add     si,ax
        mov     al,ES:XSize
        sub     si,ax
        Mov     DL,ES:YSize             ; DL := YSize
        Xor     BH,BH
        Mov     BL,ES:XSize             ; BX := XSize
        Xor     CH,CH                   ; Clear high byte of counter

ROW_YLOOP:
        mov     cl,BL                   ;CX := XSize
        REP     Movsb                   ; Copy row

        mov     al,0                    ;slap a zero on end of the row
        stosb                           ;[di] :=  0;  (di++)

        sub     si,BX                   ;move to beginning of this row
        sub     si,BX                   ; "   "      "     "  PREVIOUS row

        dec     dl
        jg      ROW_YLOOP               ;done all the rows?

        jmp     CONTINUE                ; Jump to the end of the Transformations

XYFORM:  ;----- Perform X- and Y-plane inversion  ---------------

        ;--- Start with the LAST ROW of the Image ---------

        xor     ax,ax
        mov     al,ES:ImageLen
        add     si,ax
        xor     ax,ax
        mov     al,ES:XSize
        sub     si,ax


        mov     dl,ES:YSize             ;DL-reg :=  YSize

ROW_XYLOOP:
        call    Y_XFORM_ROW             ;Xfer and Flip the row
        mov     al,0                    ;slap a zero on end of the row
        stosb                           ;[di] :=  0;  (di++)

        xor     ax,ax
        mov     al,ES:XSize
        sub     si,ax                   ;move to beginning of this row
        sub     si,ax                   ; "   "      "     "  PREVIOUS row

        dec     dl
        jg      ROW_XYLOOP              ;done all the rows?


        jmp     CONTINUE                ; Jump to the end of the Transformations



page

NOFORM:    ; ---- Perform no inversion - move only (image to WORKAREA)  ------

        Mov     ES:XBias,0              ; XBias only used on XForm
        mov     dl,ES:YSize             ; DL := YSize
        Mov     DH,ES:XSize             ; DH := XSize
        Xor     CH,CH                   ; Clear high byte of counter
        mov     al,0                    ; Clear AL for terminating row

ROW_MLOOP:
        mov     cl,DH                   ; CX := XSize
        REP     Movsb                   ; Move row
        stosb                           ; Slap a zero on end of the row

        dec     dl                      ; Decrement row counter
        jg      ROW_MLOOP               ;done all the rows?

page

CONTINUE:       ; --- After any transformations

        Pop     DS              ; Restore lost DS
        inc     XSize           ;reflects added 0-byte's added on row ends

   ;=========  Shift WORKAREA image right according to bit offset   ===========
   ;
   ;            where:  Bit Offset =  XPOS mod 8
   ;

        Mov     DI,Offset DGroup:WorkArea       ; DI -> WorkArea
        mov     ax,XPOS         ; AX =  XPOS mod 8  (Bit Offset)
        Sub     AX,XBias        ; Subtract XBias (= 0 if no XForm done)
        Mov     XPOS,AX         ; Save the adjusted XPos
        and     al,07h
        jz      NoShift         ;   (Shifting is not required)

ShiftRAgain:
        Mov     SI,DI           ; SI set to start of image
        Xor     CH,CH
        Mov     CL,ImageLen     ; How long is the image in bytes
        Add     CL,YSize        ;  + added 0 bytes
        Clc                     ; Make sure first bit shifted in is clear
ShiftRLoop:
        Rcr     byte ptr [si],1 ; Rotate WORKAREA[i] with CARRY-Bit
        Inc     si              ;          i = i + 1;
        Loop    ShiftRLoop      ; Until all bytes are shifted

        Dec     AL
        Jnz     ShiftRAgain     ; Until all bits are shifted

NoShift:
        Ret

LSetup  endp


subttl ---                                              Subroutine: CalcDot
page
;*---------------------------------------------------------------------
;                                               #####  INTERNAL  #####
; NAME          CalcDot
;
; INPUTS
;       XPOS
;       YPOS
;
; OUTPUTS
;       ES                      - Screen buffer segment for odd or even scans
;       DI                      - Screen buffer offset of the Dot's Byte
;
; DESCRIPTION
;       This procedure calculates the Display Buffer OFFSET of the
; dot addressed by the globals (XPOS,YPOS). Must not change CX.
;
;*--------------------------------------------------------------------

CalcDot Proc    Near
        Mov     AX,[YPos]               ; Get row
        Mov     ES,CGAEvenScans         ; Assume its an even row
        Test    AX,1                    ; But is it really?
        Jz      ItsEven
        Mov     ES,CGAOddScans          ; No, make it odd
ItsEven:
        Shr     AX,1                    ; Calc. # of odd/even lines
        Mul     BytesPerRow             ; Calc. offset to beginning of row
        Mov     DI,AX                   ; Save in DI
        Mov     AX,[XPos]               ; Get bit offset of column
        Shr     AX,1                    ; Calc. byte offset of column
        Shr     AX,1
        Shr     AX,1
        Add     DI,AX                   ; Store combined offset in DI
        Ret
CalcDot EndP

subttl ---                                              Sub-Routine:  SYMBOL_BLT
page
;-------------------------------------------------------------------
;                                                ###  INTERNAL  ###
;
; subroutine SYMBOL_BLT
;
;
;
;    The SYMBOL_BLT performs a BLock-Transfer of a Symbol to screen.
;
;
; INPUT
;       XSize    - The width of each row to be drawn (in bytes)
;       YSize    - The number of rows to be drawn
;       WorkArea - buffer to be written to the screen
;       DrawType - If True, Bit-Map is XOR'ed with the screen's current
;                       contents.  Otherwise it's OR'ed
;
; WARNING:  No registers are preserved.
;
;-------------------------------------------------------------------

public  SYMBOL_BLT      ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

SYMBOL_BLT      proc    near
        push    es
        push    si
        push    di

        call    CalcDot         ;figure its Display-Offset
        Mov     SI,Offset DGroup:WorkArea ; SI -> buffer to be drawn

        mov     bh,DrawType     ;Load the OR/XOR determinate flag into AH-reg
        Mov     DL,YSize        ; Load dimensions of image (# of rows)
        Mov     CL,XSize        ; (# of bytes)
        Xor     CH,CH

        Cld
SB_DrawRow:
        Lodsb                   ; Get next byte from DS:SI
        Cmp     BH,DoXor        ; Do we XOR or OR or Erase
        Jb      SB_DoOr
        Ja      SB_DoErase
        Xor     ES:[DI],AL      ; We XOR it
        Jmp     SB_DrewByte
SB_DoErase:
        Xor     AL,0FFh
        And     ES:[DI],AL      ; We Erase it
        Jmp     SB_DrewByte
SB_DoOr:
        Or      ES:[DI],AL      ; We OR it
SB_DrewByte:
        Inc     DI              ; Advance to next position in screen buffer
        Loop    SB_DrawRow

        Dec     DL              ; Decrement # of rows to draw
        Jz      SB_Done         ; Are we done?

        Mov     AX,ES
        Cmp     AX,CGAEvenScans ; Was it an even scan?
        Je      SB_SkipIncOffset
        Add     DI,BytesPerRow  ; Increase offset when going from odd to even
SB_SkipIncOffset:
        Mov     CL,XSize
        Xor     CH,CH
        Sub     DI,CX           ; Backup to start of next line
        Xor     AX,ScanToggle   ; Toggle between even and odd scan lines
        Mov     ES,AX           ; Store result
        Jmp     SB_DrawRow
SB_Done:
        pop     di      ; restore all them register saved above
        pop     si
        pop     es
        ret
SYMBOL_BLT      endp




subttl ---                                              Procedure:  Draw
page
;-------------------------------------------------------------------
;
; Procedure Draw (CharVal, Option: Byte; XPos, YPos: Integer);
;
;
;    The Draw routine draws the Image of the CharVal symbol at the
; (XPos, YPos) passed, in accordance with the Option parameter.
;
;-------------------------------------------------------------------

STARTPROC       Draw

        Call    LSetup                  ; Put image in WorkArea, set dimensions
        Mov     DrawType,DoOr           ; So SYMBOL_BLT will OR, rather than XOR
        Call    SYMBOL_BLT              ; Draw image on the screen

EndProc         Draw,8


subttl ---                                              Procedure:  Erase
page
;-------------------------------------------------------------------
;
; Procedure Erase (Code, Option: byte; X, Y: integer);
;
;
;    The Erase routine erases the Image of the symbol specified by Code at
; position (X,Y).
;
;-------------------------------------------------------------------

STARTPROC       Erase

        Call    LSetup                  ; Put image in WorkArea, set dimensions
        Mov     DrawType,DoErase        ; So SYMBOL_BLT will OR, rather than XOR
        Call    SYMBOL_BLT              ; Draw image on the screen

EndProc         Erase,8


subttl ---                                              Procedure:  QDraw
page
;-------------------------------------------------------------------
;
; procedure QDraw (Code, Option: byte; X, Y: integer; Xor: Boolean);
;
;
;    The QDraw routine saves the area of the screen to be covered by <Code>
;       and then draws <Code> in that area.
;
;-------------------------------------------------------------------

        Public  QDraw
QDraw   Proc    Far

        Pop     AX                      ; Get old IP
        Pop     BX                      ; Get old CS
        Pop     CX
        Mov     DrawType,CL             ; DrawType := Xor (True or False)
        Push    BX                      ; Restore old CS
        Push    AX                      ; Restore old IP

        Push    BP                      ; Save the caller's BP
        Mov     BP,SP                   ; Set the base pointer

                ; Now the parameter stack is the same as for Draw or Erase
        Call    LSetup                  ; Put image in WorkArea, set dimensions

        Mov     AX,XPos                 ; Save dimensions for QErase
        Mov     QD_XPos,AX
        Mov     AX,YPos
        Mov     QD_YPos,AX
        Mov     AL,XSize
        Mov     QD_XSize,AL
        Mov     AL,YSize
        Mov     QD_YSize,AL

                ; Set up source and destination pointers
        Call    CalcDot                    ; ES:DI -> Start position on screen
        Mov     SI,Offset DGroup:QSaveArea ; DS:SI -> QSaveArea

        Mov     DL,QD_YSize        ; Load dimensions of image (# of rows)
        Mov     CL,QD_XSize        ; (# of bytes per row) before DS changes
        Xor     CH,CH

        Push    DS                      ; Now switch ES:DI and DS:SI
        Mov     AX,DS                   ; Make screen the source
        Mov     BX,ES
        Mov     DS,BX
        Mov     ES,AX
        Xchg    SI,DI
        Cld

QD_MoveRow:
        Rep     Movsb           ; Move line to QSaveArea

        Dec     DL              ; Decrement # of rows to draw
        Jz      QD_Saved        ; Are we done?

        Mov     AX,DS
        Cmp     AX,ES:CGAEvenScans ; Was it an even scan?
        Je      QD_SkipIncOffset
        Add     SI,ES:BytesPerRow  ; Increase offset when going from odd to even
QD_SkipIncOffset:
        Mov     CL,ES:QD_XSize
        Xor     CH,CH
        Sub     SI,CX           ; Backup to start of next line
        Xor     AX,ScanToggle   ; Toggle between even and odd scan lines
        Mov     DS,AX           ; Store result
        Jmp     QD_MoveRow
QD_Saved:
        Pop     DS              ; Restore DS

        Call    SYMBOL_BLT              ; Draw image on the screen

EndProc         QDraw,8                 ; Xor has already been popped


subttl ---                                              Procedure:  QErase
page
;-------------------------------------------------------------------
;
; procedure QErase;
;
;
;    The QErase routine restores to the screen the area saved by the last
;       QDraw.
;
;-------------------------------------------------------------------

STARTPROC       QErase

                ; Set up source and destination pointers
        Mov     AX,QD_XPos             ; Set XPos and YPos for CalcDot
        Mov     XPos,AX
        Mov     AX,QD_YPos
        Mov     YPos,AX
        Call    CalcDot                    ; ES:DI -> Start position on screen
        Mov     SI,Offset DGroup:QSaveArea ; DS:SI -> ?SaveArea (? = Q or B)

        Mov     DL,QD_YSize        ; Load dimensions of image (# of rows)
        Mov     CL,QD_XSize        ; (# of bytes per row)
        Xor     CH,CH
        Cld

QE_MoveRow:
        Rep     Movsb           ; Move line to QSaveArea

        Dec     DL              ; Decrement # of rows to draw
        Jz      QE_Done        ; Are we done?

        Mov     AX,ES
        Cmp     AX,CGAEvenScans ; Was it an even scan?
        Je      QE_SkipIncOffset
        Add     DI,BytesPerRow  ; Increase offset when going from odd to even
QE_SkipIncOffset:
        Mov     CL,QD_XSize
        Xor     CH,CH
        Sub     DI,CX           ; Backup to start of next line
        Xor     AX,ScanToggle   ; Toggle between even and odd scan lines
        Mov     ES,AX           ; Store result
        Jmp     QE_MoveRow

QE_Done:

EndProc         QErase



subttl ---                                              Procedure:  SaveBoxArea
page
;-------------------------------------------------------------------
;
; procedure SaveBoxArea (Save: Boolean);
;
;
;    The SaveBoxArea routine saves or restores the keybox screen area.
;       Save  -- If Save = True, the box is saved; otherwise it's restored.
;
;-------------------------------------------------------------------
SBA_Save        Equ     6       ; Offset from BP to the parameter

STARTPROC       SaveBoxArea

        Push    DS                      ; Save DS
        Cld

        Test    Byte Ptr SBA_Save[BP],True ; Are we saving or restoring
        Jz      SBA_Restore

        Mov     AX,DS                   ; We're saving
        Mov     ES,AX
        Mov     DI,Offset DGroup:BSaveArea ; ES:DI -> BSaveArea
        Mov     DS,CGAEvenScans
        Mov     SI,BoxStart             ; DS:SI -> Box area of screen
        Jmp     SBA_Go

SBA_Restore:
        Mov     ES,CGAEvenScans         ; We're restoring
        Mov     DI,BoxStart             ; ES:DI -> Box area of screen
        Mov     SI,Offset DGroup:BSaveArea ; DS:SI -> BSaveArea

SBA_Go:
        Mov     CX,BSaveBytes/2
        Rep     Movsb                   ; Move even scan lines
        Jz      SBA_2Restore            ; Flags from old Test are still valid

        Mov     DS,ES:CGAOddScans
        Mov     SI,BoxStart             ; Reset DS:SI, this time to odd scans
        Jmp     SBA_2Go

SBA_2Restore:
        Mov     ES,CGAOddScans
        Mov     DI,BoxStart             ; Reset ES:DI, this time to odd scans

SBA_2Go:
        Mov     CX,BSaveBytes/2
        Rep     Movsb                   ; Move odd scan lines

        Pop     DS                      ; Restore DS

EndProc         SaveBoxArea,2


subttl ---                                              Procedure:  ScanSym
page
;-------------------------------------------------------------------
;
;
; Function ScanSym (Code, Option: byte; Buffer: AdrMem; Index: Integer):
;                                                               Integer;
;
;
;    The ScanSym routine returns the offset from Index of the 1st occurrence,
; after Index, of the symbol specified by Code.
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------
SSCode          Equ     12              ; Symbol code
SSOption        Equ     10              ; Orientation and Flop
SSBuffer        Equ     8               ; Buffer containing file being edited
SSIndex         Equ     6               ; Index into Buffer to start search from


public  ScanSym         ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

STARTPROC       ScanSym

        Push    DS                      ; Save DS
        Mov     SI,SSBuffer[BP]         ; DS:SI -> position in buffer
        Add     SI,SSIndex[BP]
        Mov     BL,SSCode[BP]           ; BX holds what we're looking for
        Mov     BH,SSOption[BP]
        Mov     DX,SI                   ; Save start position in DX
        Cld                             ; Search in + direction

SS05:   Lodsw
        Cmp     AX,BX                   ; Is this it?
        Je      SSDone
        Cmp     AL,EOF                  ; Did we reach the end of the buffer?
        Je      SSDone                  ; EOF matches everything.
        Cmp     AL,NuSign               ; Was it really a control code?
        Jae     SS05
        Lodsw                           ; No, skip next word
        Jmp     Short SS05              ; Go back and try again

SSDone: Mov     AX,SI                   ; Put current offset in AX
        Sub     AX,2                    ; 'Undo' last Lodsw
        Sub     AX,DX                   ; Subtract start pos to get offset

        Pop     DS                      ; Restore DS
ENDPROC ScanSym,8


subttl ---                                              Procedure:  ScanCode
page
;-------------------------------------------------------------------
;
;
; Function ScanCode (StopSigns : Boolean;
;                    Dir       : Integer;
;                    Buffer    : AdrMem;
;                    Index     : Integer) : Integer;
;
;
;    The ScanCode routine starts at Buffer + Index and goes in the direction
; specified in Dir, looking for a code symbol.
;
; StopSigns : If true, stop and NewSign or NewLine, If false stop only at NewLine.
; Dir       : 1 => search in + direction; -1 => search in - direction.
; Buffer    : Address of buffer being modified.
; Index     : Position within buffer to start search from.
; ScanCode  : Offset from Buffer + Index to first code found.
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------
SCStopSigns     Equ     12              ; Symbol code
SCDir           Equ     10              ; Orientation and Flop
SCBuffer        Equ     8               ; Buffer containing file being edited
SCIndex         Equ     6               ; Index into Buffer to start search from


public  ScanCode        ;=-=-=-=-=-=-=-  DEBUG Only   -=-=-=-=-=-=-=

STARTPROC       ScanCode

SC02:   Mov     BL,NuSign               ; Lowest valued code in BX
        Cmp     Byte Ptr SCStopSigns[BP],True ; Do we stop at NewSign?
        Je      SC05
        Inc     BL                      ; No, NewSign + 1 = NewLine
SC05:   Mov     SI,SCBuffer[BP]         ; DS:SI -> start position
        Mov     CX,SI                   ; Save start of buffer
        Add     SI,SCIndex[BP]

        Cmp     Byte Ptr SCDir[BP],0    ; What direction do we search in?
        Jg      SC08
        Std                             ; - direction
        Lodsw                           ; Skip the first code
        Jmp     Short SC09
SC08:   Cld                             ; + direction
        Lodsw                           ; Get the first char, maybe skip it
        Cmp     AL,EOF                  ; Is it an EOF?
        Mov     AX,0                    ; return distance of 0
        Je      SCDone                  ; If so and counting up, then exit.

SC09:   Mov     DX,SI                   ; Save start position
SC10:   Lodsw                           ; Get next word
        Cmp     SI,CX
        Jb      SC15                    ; Are we before the start of the buffer
        Cmp     AL,EOF                  ; Did we reach the end of the file?
        Je      SC15
        Cmp     AL,BL                   ; Is it what we want
        Jb      SC10

SC15:   Mov     AX,SI                   ; Get current offset in AX
        Sub     AX,DX                   ; Subtract start position to get difference
SCDone:

ENDPROC ScanCode,8


subttl ---                                              Procedure:  DrawLine
page
;-------------------------------------------------------------------
;
;
; procedure DrawLine (X1, Y1, X2, Y2, BitVal: Integer)
;
;
; DrawLine - Draw a vertical or horizontal line from (0-relative) screen
;            dot coordinate (X1,Y1) to (X2,Y2).  BitVal = 1 means a solid
;            line, BitVal = 2 means a dashed (every other bit is on) line.
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------

DL_BitVal       Equ     6
DL_Y2           Equ     8
DL_X2           Equ     10
DL_Y1           Equ     12
DL_X1           Equ     14

StartProc       DrawLine

        Cld                             ; Make sure all string ops go up
        Mov     AX,DL_X1[BP]            ; Get X1
        Cmp     AX,DL_X2[BP]            ; Compare with X2
        Je      DL_Vertical
        Jl      DL_Horizontal

        Xchg    AX,DL_X2[BP]            ; Its horizontal but backwards
        Mov     DL_X1[BP],AX            ; Switch the end points
        Mov     AX,DL_Y1[BP]
        Xchg    AX,DL_Y2[BP]
        Mov     DL_Y1[BP],AX

DL_Horizontal:
        Mov     AX,DL_X1[BP]
        Mov     XPos,AX
        Mov     AX,DL_Y1[BP]
        Mov     YPos,AX
        Call    CalcDot                 ; ES:DI -> start dot in screen buffer

        Mov     BX,XPos
        And     BX,07H                  ; XPos Mod 8
        Mov     AL,ES:[DI]              ; See what's already on the screen
        And     AL,DL_NotMasks[BX]      ; Clear part that needs to be written

        Xor     DH,DH
        Mov     DL,DL_BitVal[BP]        ; What kind of line are we writing?
        Shl     DL,1                    ; Multiply by the length of each
        Shl     DL,1                    ;  so that DL will index from the
        Shl     DL,1                    ;  start of all the tables to the
                                        ;  beginning of the desired table
        Add     BX,DX                   ; BX -> correct entry of correct table
        Or      AL,DL_Masks[BX]         ; Get the beginning of the line

        Inc     Word Ptr DL_X2[BP]      ; Point to bit past last bit drawn
        Mov     CX,DL_X2[BP]            ; Get the end point in bits
        Shr     CX,1                    ; Convert to bytes
        Shr     CX,1
        Shr     CX,1
        Mov     BX,XPos                 ; Get start point in bits
        Shr     BX,1                    ; Convert to bytes
        Shr     BX,1
        Shr     BX,1
        Sub     CX,BX                   ; CX gets number of bytes to write
        Jcxz    DL_LastByte             ; If only one byte, make it last
        Dec     CX                      ; - 2 end bytes, + 1 since its inclusive

        Stosb                           ; Now write first byte out.

        Mov     BX,DX                   ; Get beginning of correct table
        Mov     AL,DL_Masks[BX]         ; First entry is allways a full byte
        Rep     Stosb                   ; Write the middle bytes out

DL_LastByte:
        Mov     BX,DL_X2[BP]            ; Get end point in bits
        And     BX,07h                  ; X2 mod 8
        And     AL,DL_NotMasks[BX]      ; Clear part of line that won't be used
        Mov     AH,ES:[DI]              ; Get what's currently on the screen
        And     AH,DL_SolidMasks[BX]    ; Clear part that will be overwritten
        Or      AL,AH                   ; Move what was there onto the end of AL
        Stosb                           ; Write whole thing back to the screen

        Jmp     DL_Done                 ; Ahhhh!

DL_Vertical:

        Mov     AX,DL_Y1[BP]            ; Get X1
        Cmp     AX,DL_Y2[BP]            ; Compare with X2
        Jle     DL_SkipSwitch

        Xchg    AX,DL_Y2[BP]            ; Its vertical but upside-down
        Mov     DL_Y1[BP],AX            ; Switch the end points
        Mov     AX,DL_X1[BP]
        Xchg    AX,DL_X2[BP]
        Mov     DL_X1[BP],AX

DL_SkipSwitch:
        Mov     AX,DL_X1[BP]
        Mov     XPos,AX
        Mov     AX,DL_Y1[BP]
        Mov     YPos,AX
        Call    CalcDot                 ; Set up ES:DI

        Mov     CX,XPos
        And     CL,07h                  ; Get offset into byte
        Mov     AL,80H                  ; Get ready to find the bit.
        Ror     AL,CL                   ; Keep shifting til we get there
        Mov     BL,AL
        Not     BL                      ; BL is used to draw clear bits

        Mov     AH,DL_BitVal[BP]        ; What kind of line is this
        Mov     BH,2                    ; Count for how many times (odd and even lines)
DL_VLoop:
        Mov     CX,DL_Y2[BP]            ; Get the end point in bits
        Sub     CX,DL_Y1[BP]            ; CX gets number of bits to write
        Add     CL,BH   ; (CX += BH) takes care of even vs. odd number of lines
        Adc     CH,0    ;  and inclusion of both end points.  Think about it.
        Shr     CX,1                    ; Div 2 for even or odd lines
        Jcxz    DL_Done                 ; Will only happen 2nd time through
        Cmp     AH,0                    ; > 0, do solid bits
        Jg      DL_VSolid

        Mov     AL,BL                   ; Prepare to draw a clear bits
DL_VClear:
        And     ES:[DI],AL              ; Clear the line
        Add     DI,BytesPerRow          ; Move to next line
        Loop    DL_VClear
        Jmp     DL_AgainTest

DL_VSolid:
        Or      ES:[DI],AL              ; Draw solid line
        Add     DI,BytesPerRow          ; Move to next line
        Loop    DL_VSolid

DL_AgainTest:
        Dec     BH
        Jz      DL_Done
        Ror     AH,1                    ; Clear stays clear, solid stays solid
        Shr     AH,1                    ;  but dotted goes to clear
        Inc     YPos                    ; Start one line down
        Mov     CX,AX                   ; Save AX since CalcDot munches it
        Call    CalcDot
        Mov     AX,CX                   ; Restore CalcDot
        Jmp     DL_VLoop                 ; Do it again

DL_Done:

EndProc         DrawLine,10

subttl ---                                              Procedure:  EraseLine
page
;-------------------------------------------------------------------
;
;
; procedure EraseLine(LNum: LnX);
;
;
; EraseLine - Erases a line of text.
;
; Parameters:
;       LNum : The number of the line to be erased (0..2)
;
;-------------------------------------------------------------------
;
; ---  Parameter Offsets -----------

EL_LNum         Equ     6

StartProc       EraseLine

        Mov     BL,EL_LNum[BP]
        Xor     BH,BH                   ; LNum is a byte value
        Shl     BX,1                    ; Convert to word index
        Mov     ES,CGAEvenScans
        Mov     DI,LineEAddrs[BX]       ; Address of even scans to be erased
        Mov     CX,LineSize             ; Get count of bytes to be cleared
        Xor     AL,AL                   ; Clear AL for writing
        Cld
        Rep     Stosb                   ; Erase even scans
        Mov     ES,CGAOddScans
        Mov     DI,LineOAddrs[BX]       ; Address of odd scans to be erased
        Mov     CX,LineSize             ; Get count of bytes to be cleared
        Rep     Stosb                   ; Erase odd scans

EndProc         EraseLine,2

subttl ---                                              Procedure:  LMoveLine
page
;-------------------------------------------------------------------
;                                                 ### INTERNAL ###
;
; procedure LMoveLine;
;
;
; LMoveLine - Moves text from one line to another.
;
; Input:
;       SI : Holds line number of source line
;       DI : Holds line number of destination line
;
; CAUTION: DS IS DESTROYED!
;
;-------------------------------------------------------------------

LMoveLine       Proc    Near

        Push    DS                      ; Save DS
        Shl     SI,1                    ; Convert SI and DI to Word indexes
        Shl     DI,1
        Push    SI                      ; Save these for the odd scans
        Push    DI
        Push    DS

        Mov     ES,CGAEvenScans
        Mov     DI,LineEAddrs[DI]
        Mov     SI,LineOAddrs[SI]
        Mov     DS,CGAOddScans
        Mov     CX,LineSize
        Cld
        Rep     Movsb                   ; Copy even scans

        Pop     DS                      ; Restore to get odd scans
        Pop     DI
        Pop     SI

        Mov     ES,CGAOddScans
        Mov     DI,LineOAddrs[DI]
        Mov     SI,LineEAddrs[SI]
        Mov     DS,CGAEvenScans
        Mov     CX,LineSize
        Rep     Movsb                   ; Copy odd scans

        Pop     DS
        Ret
LMoveLine       EndP


subttl ---                                              Procedure:  ScrollUp
page
;-------------------------------------------------------------------
;
;
; procedure ScrollUp;
;
;
; ScrollUp - Scrolls all three text lines up one line.  The last line is
;       left empty.
;
;
;-------------------------------------------------------------------

StartProc       ScrollUp

        Push    DS                      ; Save DS

        Mov     SI,1                    ; Set up arguments for LMoveLine
        Mov     DI,0
        Call    LMoveLine               ; Move line 1 to line 0

        Mov     SI,2                    ; Set up arguments for LMoveLine
        Mov     DI,1
        Call    LMoveLine               ; Move line 2 to line 1

        Mov     BX,2                    ; Push param for EraseLine to erase
        Push    BX                      ;  the last line.
        Call    EraseLine               ; Erase line 2

        Pop     DS

EndProc         ScrollUp


subttl ---                                              Procedure:  ScrollDown
page
;-------------------------------------------------------------------
;
;
; procedure ScrollDown (LNum: LnX);
;
;
; ScrollDown - Scrolls down each line of text starting from LNum.  Line LNum
;       is erased.
;
; Parameters:
;       LNum : Line number to start scrolling from
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----

SD_LNum         Equ     6

StartProc       ScrollDown

        Push    DS                      ; Save DS

        Mov     DL,SD_LNum[BP]          ; Get the line number
        Xor     DH,DH

        Mov     SI,1                    ; Set up arguments for LMoveLine
        Mov     DI,2
        Cmp     DI,DX                   ; Is this the line to erase
        Jle     SD_DoErase
        Call    LMoveLine               ; No, move line 1 to line 2

        Mov     SI,0                    ; Set up arguments for LMoveLine
        Mov     DI,1
        Cmp     DI,DX                   ; Is this the line to erase
        Jle     SD_DoErase
        Call    LMoveLine               ; No, move line 0 to line 1

        Xor     DI,DI                   ; Then we are going to erase line 0

SD_DoErase:
        Push    DI                      ; DI holds number to erase
        Call    EraseLine               ; So erase it.

        Pop     DS

EndProc         ScrollDown,2


subttl ---                                              Procedure:  EpsonPrep
page
;-------------------------------------------------------------------
;
;
; procedure EpsonPrep (ScrLine: ScrInx;
;                      Var PrintBuf: AdrMem;
;                      Var NonBlank: Boolean);
;
;       EpsonPrep converts screen graphics to Epson printing format
;
; Parameters:
;       ScrLine  : starting line of 4 screen lines to convert
;       PrintBuf : destination of converted data (640 bytes)
;       NonBlank : returns TRUE if converted lines were not blank
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----

EPScrLine       Equ     10
EPPrintBuf      Equ     8
EPNonBlank      Equ     6

StartProc       EpsonPrep
        Push    DS

        Mov     XPos,0
        Mov     AX,EPScrLine[BP]
        Mov     YPos,AX
        Call    CalcDot

        Mov     AX,DS
        Mov     BX,ES
        Mov     ES,AX
        Mov     DS,BX
        Mov     SI,DI

        Cld                             ; Screen ptr counts up.
        Mov     DL,False                ; Assume blank initially
        Mov     BH,4                    ; Store 4 rows for printing
EPNextRow:
        Mov     DI,EPPrintBuf[BP]       ; Reset DI to start of buf for each row
        Mov     BL,80                   ; Load count for # of bytes/row
EPMapRow:
        Mov     CX,8                    ; Load count for # of bits/byte
        Lodsb                           ; Get next byte from the screen
        Or      AL,AL                   ; Was anything in it?
        Jz      EPMapByte
        Mov     DL,True                 ; Yes, set NonBlank to True
EPMapByte:
        Rcl     AL,1                    ; Get next bit of the screen
        Jc      EPDoDot                 ; Was the bit clear or set?
        Rcl     Byte Ptr ES:[DI],1      ; Clear, store the bit once
        Clc                             ; Duplicate the bit
        Rcl     Byte Ptr ES:[DI],1      ; Store the bit twice
        Jmp     EPDotDone
EPDoDot:
        Rcl     Byte Ptr ES:[DI],1      ; Bit set, store the bit once
        Stc                             ; Duplicate the bit
        Rcl     Byte Ptr ES:[DI],1      ; Store the bit twice
EPDotDone:
        Inc     DI                      ; Store next screen bit in next byte
        Loop    EPMapByte               ; Are we done with the byte?
        Dec     BL
        Jnz     EPMapRow                ; Are we done with the row?
        Dec     BH
        Jz      EPDone                  ; Are we done?

        Mov     AX,DS
        Cmp     AX,ES:CGAOddScans   ; Was it an even scan?
        Je      EP_SkipDecOffset
        Sub     SI,80               ; Reset offset when going from even to odd
EP_SkipDecOffset:
        Xor     AX,ScanToggle       ; Toggle between even and odd scan lines
        Mov     DS,AX               ; Store result
        Jmp     EPNextRow
        
EPDone:
        Pop     DS
        Mov     BX,EPNonBlank[BP]       ; Get pointer to NonBlank
        Mov     [BX],DL                 ; And store result

EndProc         EpsonPrep,6


subttl ---                                              Procedure:  GrafPrep
page
;-------------------------------------------------------------------
;
;
; procedure GrafPrep (ScrLine: ScrInx;
;                     Var GrafBuf: AdrMem);
;
;       GrafPrep moves screen graphics data to a string variable.
;       It is used to convert screen graphics to PCX file format
;       and to HP LaserJet printing format.
;
; Parameters:
;       ScrLine  : screen line to convert
;       GrafBuf  : destination of converted data
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----

GPScrLine       Equ     8
GPGrafBuf       Equ     6

StartProc       GrafPrep
        Push    DS

        Mov     XPos,0
        Mov     AX,GPScrLine[BP]
        Mov     YPos,AX
        Call    CalcDot

        Mov     AX,DS
        Mov     BX,ES
        Mov     ES,AX
        Mov     DS,BX
        Mov     SI,DI

        Cld                             ; Screen ptr counts up.
        Mov	CX,80			; 80 bytes per screen line
        Mov     DI,GPGrafBuf[BP]        ; Set DI to start of string

GPMoveLn:
        Lodsb                           ; Get byte from screen
        Mov     Byte Ptr ES:[DI],AL     ; Store it in string
	Inc	DI
        Loop    GPMoveLn                ; do the next one
        
GPDone:
        Pop     DS

EndProc         GrafPrep,4


subttl ---                                              Procedure: GetDTA
page
;-------------------------------------------------------------------
;
;
; Function GetDTA: AdsMem;
;
;
; GetDTA - Gets the current DTA address
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----

StartProc       GetDTA

        Mov     AH,DOSGetDTA
        Int     DOSFunction
        Mov     DX,ES
        Mov     AX,BX

EndProc         GetDTA



subttl ---                                              Procedure:  AGotChar
page
;-------------------------------------------------------------------
;
;
; procedure AGotChar: Boolean
;
;
; AGotChar - Returns true if a character is available.
;
;-------------------------------------------------------------------

StartProc       AGotChar

        Xor     AX,AX                   ; Assume false
        Mov     DI,NextCh               ; Get index of next character
        Inc     DI                      ; NextCh -> before the start of the buffer
        And     DI,CBuffSize - 1        ;  - Increment mod CBuffSize
        Cmp     DI,LastCh               ; Is there anything in the buffer
        Je      AGCFalse                ; If not, return false
        Inc     AX                      ; Yes, return true;
AGCFalse:

EndProc         AGotChar


subttl ---                                              Procedure:  AGetChar
page
;-------------------------------------------------------------------
;
;
; procedure AGetChar (Var Ch: Char; Var Opt: Boolean);
;
;
; AGetChar - Gets the latest char typed in on the keyboard.
;
; Parameters:
;       Ch - Returned with the ASCII code of the character typed,
;               if the CTRL key was hit, the return code will be between
;               0 and 31.
;       Opt - Returns true of the Alt key was pressed.
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----

AGC_Ch          Equ     8
AGC_Opt         Equ     6

StartProc       AGetChar
; zombie:out:start
;;        Xor     AH,AH
;;        Int     KBDIOInt                ; Get next char in buffer
;;        Mov     CX,AX                   ; Save the code in CX
;;        Mov     AH,2
;;        Int     KBDIOInt                ; Quick, get the shift state before
;;                                        ; -it changes.
;;
;;        Mov     BX,LastCh
;;        Mov     CharBuff[BX],CL         ; Store the char
;;        Mov     SpecBuff[BX],CH         ; Store the special code
;;        Mov     ShiftBuff[BX],AL        ; Store the shift state
;;        Inc     BX                      ; Yes, store the char info
;;        And     BX,CBuffSize - 1
;;        Mov     LastCh,BX
;;
; zombie:out:end

        Mov     DI,NextCh               ; Get index of next character
        Inc     DI                      ; NextCh -> before the start of the buffer
        And     DI,CBuffSize - 1       ;  - Increment mod CBuffSize
AGCWait:
        Cmp     DI,LastCh               ; Is there anything in the buffer
        Je      AGCWait                 ; If not, check again

        Mov     DH,ShiftBuff[DI]        ; Get the shift state of the key
        And     DH,AltFlag              ; Isolate the Alt shift state flag
        Mov     CL,3
        Shr     DH,CL                   ; Move it to bit 0

        Mov     DL,CharBuff[DI]         ; Get the character
        Or      DL,DL                   ; Is it a special character
        Jz      AGCSpec
        Cmp     DL,26                   ; Is it a Ctrl key
        Ja      AGCDone

        Mov     AL,ShiftBuff[DI]        ; Yes, get the shift state again
        Mov     AH,AL                   ; Duplicate it
        And     AL,RShiftFlag           ; Isolate the shift flags
        And     AH,LShiftFlag
        Shr     AH,1                    ; Align the flags to bit 1 of AL and AH
        Or      AL,AH                   ; AL = R-shift Or L-shift
        Or      DH,AL                   ; Opt = R=shift Or L-shift Or Alt
        Jmp     AGCDone                 ; Outta-here

AGCSpec:
        Mov     BL,SpecBuff[DI]         ; Yes, get the special char code
        Xor     BH,BH
        Mov     DL,CharMap[BX]          ; Yes, convert to the appropriate code
        Mov     BL,ShiftBuff[DI]        ; What about the control key
        Test    BL,CtrlFlag             ; Was <ctrl> hit?
        Jz      AGCTAlt
        And     DL,1Fh                  ; Yes, clear high 3 bits.
AGCTAlt:
        Test    DL,KAltFlag             ; Should we simulate the AltFlag
        Jz      AGCDone
        And     DL,Not KAltFlag         ; Clear the KAltFLag
        Mov     DH,True                 ; Yes

AGCDone:
        Mov     BX,AGC_Opt[BP]          ; Get address of option parameter
        Mov     [BX],DH                 ; Store the Alt state
        Mov     BX,AGC_Ch[BP]           ; Get address of the char parameter
        Mov     [BX],DL                 ; Store the character
        Mov     NextCh,DI               ; Store updated pointer

EndProc         AGetChar,4


subttl ---                                              Procedure:  RenameFile
page
;-------------------------------------------------------------------
;
; procedure RenameFile (Const oldname, newname: LString);
;
; Rename a file from oldname to newname
;
; Parameters:
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
RFOldName       Equ     10                      ; Skip the length
RFNewName       Equ     6

StartProc       RenameFile

        Mov     AX,DS
        Mov     ES,AX
        Mov     DX,RFOldName[BP]
        Inc     DX                              ; Get past the length byte
        Mov     DI,RFNewName[BP]
        Inc     DI
        Mov     AH,DOSRenameFile
        Int     DOSFunction

EndProc         RenameFile,8


subttl ---                                              Procedure:  GetCurPath
page
;-------------------------------------------------------------------
;
; procedure GetCurPath (Var Patholdname: LString);
;
; Get the current disk and drive.
;
; Parameters:
;       LString : returns with the result
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
GCPPathLen      Equ     8
GCPPath         Equ     6

StartProc       GetCurPath

        Mov     AH,DOSCurDisk           ; Get the current disk
        Int     DOSFunction

        Mov     SI,GCPPath[BP]          ; Find the place to put the result
        Inc     SI                      ; Skip the length byte
        Add     AL,'A'                  ; Convert drive type to drive letter
        Mov     [SI],AL                 ; Store the drive letter
        Inc     SI
        Mov     [SI],Byte Ptr ':'       ; Store some more stuff
        Inc     SI
        Mov     [SI],Byte Ptr '\'
        Inc     SI

        Mov     AH,DOSGetCurDir         ; Now get the current directory
        Xor     DL,DL                   ; Use the default disk
        Int     DOSFunction

        Jnb     GCP01                   ; current directory ok?
        
        Mov     BX,GCPPath[BP]          ;   no - return empty string
        Mov     AL,0
        Mov     [BX],AL
        Jmp     GCP02

GCP01:  Mov     CX,GCPPathLen[BP]       ; Get the max length of the result
        Mov     BX,GCPPath[BP]          ; Get a pointer to the result
        Mov     DI,BX                   ; Get another pointer to the result
        Inc     DI                      ; DI points past length byte
        Mov     AX,DS
        Mov     ES,AX
        Xor     AL,AL
        Cld
        Repnz   Scasb                   ; Look for the 0 terminator
        Sub     DI,BX                   ; Subtract the beginning of the string
        Dec     DI                      ; DI was two past the end of the string
        Dec     DI
        Mov     AX,DI                   ; We only want the low order byte
        Mov     [BX],AL                 ; Store the length of the result
GCP02:

EndProc         GetCurPath,4


subttl ---                                              Procedure:  GotoXY
page
;-------------------------------------------------------------------
;
; Procedure GotoXY (X, Y: Integer);
;
; Position the text cursor on column X, row Y (0-based)
;
; Parameters:
;       X : text column (0 through 79)
;       Y : text row (0 through 24)
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
GXYColumn       Equ     8
GXYRow          Equ     6

StartProc       GotoXY

        Mov     DH,GXYRow[BP]           ; Get row coordinate
        Mov     DL,GXYColumn[BP]        ; Get column coordinate
        Xor     BH,BH                   ; Setup for page 0 only
        Mov     AH,SetCursor            ; Actually position cursor
        Int     VideoInt

EndProc         GotoXY,4


subttl ---                                              Procedure:  EdBitSet
page
;-------------------------------------------------------------------
;
; Procedure EdBitSet (XBit, Y, BitVal: Integer);
;
; Set the bit at (XBit, Y) to BitVal
;
; Parameters:
;       XBit   : zero-based column of bit
;       Y      : zero-based row of bit 
;       BitVal : value of bit to set
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
EBSColumn       Equ     10
EBSRow          Equ     8
EBSBitVal       Equ     6

StartProc       EdBitSet

        Mov     CX,EBSColumn[BP]        ; Get column value
        Mov     XPOS,CX  
        Mov     AX,EBSRow[BP]           ; Get row value
        Mov     YPOS,AX
        Call    CalcDot                 ; Get address of byte w/ dot
        And     CL,07H                  ; Get bit rotate count
        Mov     AL,80h                  ; Get on-bit value at low pixel
        Mov     AH,EBSBitVal[BP]        ; Get new bit value
        Ror     AH,1                    ; Position at low pixel
        Shr     AX,CL                   ; Move bits into place
        Not     AL                      ; Get bit mask
        And     ES:[DI],AL              ; Wipe out old victim bit
        Or      ES:[DI],AH              ; Replace with new bit

EndProc         EdBitSet,6


subttl ---                                              Procedure:  EdBlkSet
page
;-------------------------------------------------------------------
;
; Procedure EdBlkSet (XBit, Y, BitVal: Integer);
;
; Draw a 3x3 block whose upper left corner is (XBit, Y).  The block will
; be empty if BitVal = 0, and full if BitVal = 1.
;
; Parameters:
;       XBit   : zero-based column of bit
;       Y      : zero-based row of bit 
;       BitVal : value of bit to set
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
EBlColumn       Equ     10
EBlRow          Equ     8
EBlBitVal       Equ     6

StartProc       EdBlkSet

        Mov     CX,BlockSize            ; Get number of pixels in block
        Mov     AX,EBlColumn[BP]        ; Get column of left block border
        Add     AX,CX                   ; Calculate right border column
        Dec     AX

EBl01:  Push    AX                      ; Save right column border
        Push    CX                      ; Save count of rows in block

        Push    EBlColumn[BP]           ; Pass left coordinate of border
        Push    EBlRow[BP]
        Push    AX                      ; Pass right coordinate of border
        Push    EBlRow[BP]
        Push    EBlBitVal[BP]           ; Pass value of bit to draw
        Call    DrawLine                ; Draw horizontal line of bit
        
        Pop     CX                      ; Get count of rows left
        Pop     AX                      ; Get right column of border
        Inc     Word Ptr EBlRow[BP]     ; Move to next display row
        Loop    EBl01                   ; Keep going until rows complete

EndProc         EdBlkSet,6


subttl ---                                              Procedure:  EdBitDraw
page
;-------------------------------------------------------------------
;
; Procedure EdBitDraw (XByte, Y, XImageByteSize, YSize: Integer;
;                      CONSTS Im: AdsMem);
;
; Draw the image Im at (XByte * 8, Y).  XImageByteSize is the number of 
; 8 bit bytes to draw on a row, and YSize is the number of rows.
;
; Parameters:
;       XByte          : zero-based byte number within row to start
;       Y              : zero-based row number to start
;       XImageByteSize : number of 8-bit bytes to draw
;       YSize          : number of rows to draw
;       Im             : address of byte image to draw
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
EBDColumn       Equ     16
EBDRow          Equ     14
EBDImageSize    Equ     12
EBDRowSize      Equ     10
EBDIm           Equ     6

StartProc       EdBitDraw

        Cld                             ; Make memory moves go forward
        Mov     AX,EBDColumn[BP]        ; Get byte offset from start of row
        Shl     AX,1                    ; Multiply times 8 bits per byte
        Shl     AX,1
        Shl     AX,1 
        Mov     XPOS,AX                 ; Save in column position
        Mov     AX,EBDRow[BP]           ; Get starting row
        Mov     YPOS,AX                 ; Save in row position

        Mov     AX,EBDImageSize[BP]     ; Get number of 8 bit bytes in image
        Shl     AX,1                    ; Multiply by 8 to get bit count
        Shl     AX,1
        Shl     AX,1
        Mov     EBDImageSize[BP],AX     ; Save bit count for later

EBD52:  Call    CalcDot                 ; Get address of byte w/ dot
        Mov     BX,EBDImageSize[BP]     ; Get number of bits to move
        Mov     CX,BX                   ; Calculate number of bytes
        Shr     CX,1
        Shr     CX,1
        Shr     CX,1
        Mov     DX,DS                   ; Save program data segment
        Lds     SI,EBDIm[BP]            ; Get address of source image
        Rep Movsb                       ; Move however many whole bytes
        Mov     DS,DX                   ; Retrieve program data segment
        And     BX,7                    ; Get number of leftover bits
        Jz      EBD51                   ; Leftovers again, mom!?
        Lodsb                           ; Get last (fractional) byte
        Mov     AH,DL_NotMasks[BX]      ; Yes, mask off all unwanted bits
        And     AL,AH
        Not     AH
        And     AH,ES:[DI]              ; Clear out replace bits in dest
        Or      AL,AH                   ; Set new bits into dest
        Stosb                           ; Save result in destination
EBD51:  Mov     Word Ptr EBDIm[BP],SI   ; Update source for next row draw
        Inc     YPOS                    ; Move to next row
        Dec     Word Ptr EBDRowSize[BP] ; Count down row      
        Jnz     EBD52                   ; Are there any rows left?

EndProc         EdBitDraw,12


subttl ---                                              Procedure:  EdBitDraw
page
;-------------------------------------------------------------------
;
; Procedure EdBlkDraw (XByte, Y, XImageByteSize, YSize: Integer; 
;                      CONSTS Im: AdsMem);
;
; Draw the image Im at (XByte * 8, Y).  XImageByteSize is the number of 
; 8 bit bytes to draw on a row, and YSize is the number of rows.  In this
; routine, a pixel is zoomed to a 3x3 pixel block with a 1 pixel border.
;
; Parameters:
;       XByte          : zero-based byte number within row to start
;       Y              : zero-based row number to start
;       XImageByteSize : number of 8-bit bytes to draw
;       YSize          : number of rows to draw
;       Im             : address of byte image to draw
;
;-------------------------------------------------------------------

;---- Parameter Offsets ----
EBXColumn       Equ     16
EBXRow          Equ     14
EBXImageSize    Equ     12
EBXRowSize      Equ     10
EBXImage        Equ     6

StartProc       EdBlkDraw

        Mov     AX,EBXColumn[BP]        ; Get byte offset from start of row
        Shl     AX,1                    ; Multiply times 8 bits per byte
        Shl     AX,1
        Shl     AX,1 
        Mov     EBXColumn[BP],AX        ; Save real column position

        Mov     AX,EBXImageSize[BP]     ; Get number of 8 bit bytes in image
        Shl     AX,1
        Shl     AX,1
        Shl     AX,1
        Mov     EBXImageSize[BP],AX     ; Save bit count for later
        Les     DI,EBXImage[BP]         ; Get address of image start

EBX53:  Mov     CX,EBXImageSize[BP]     ; Get number of columns in image
        Mov     BX,EBXColumn[BP]        ; Get starting column
        Mov     DX,80H                  ; Get mask for first bit

EBX50:  Push    BX                      ; Save everything needed in loop
        Push    CX
        Push    DX
        Push    ES
        Push    DI

        Push    BX                      ; Pass upper left corner column
        Push    EBXRow[BP]              ; Pass upper left corner row
        Mov     AL,DL                   ; Get mask for current pixel
        And     AL,ES:[DI]              ; Determine whether set in image  
        Jz      EBX51
        Mov     AL,1                    ; If so, set bit on
EBX51:  Xor     AH,AH                   ; Clear out junk byte
        Push    AX                      ; Pass bit value
        Call    EdBlkSet                ; Go draw bit box

        Pop     DI                      ; Restore pointer into image
        Pop     ES
        Pop     DX                      ; Restore bit mask for image
        Pop     CX                      ; Restore count of columns left
        Pop     BX                      ; Restore current column number

        Mov     AX,DX                   ; Move to next image byte only if
        And     AL,1                    ;  ... image mask indicates 8 bits
        Add     DI,AX                   ;  ... used in current image byte
        Ror     DL,1                    ; Move mask to next bit
EBX52:  Add     BX,BlockSize + BlockPad ; Move to position for next bit block
        Loop    EBX50                   ; Loop for number of bits in row

        Add     Word Ptr EBXRow[BP],BlockSize + BlockPad
        Cmp     DL,80H                  ; Determine if just moved byte pointer
        Je      EBX54
        Inc     DI                      ; No, must move to next image byte
EBX54:  Dec     Word Ptr EBXRowSize[BP] ; Are there any rows left?
        Jnz     EBX53                   ; Continue if so

EndProc         EdBlkDraw,12


SW_Supp ends      ; ---- End of the Code Segment  ---------

        end

